package ai.queens;

import java.util.Arrays;

import ai.App;
import ai.Resolver;
import ai.metaheuristic.genetic.GeneticAlgorithm;
import ai.metaheuristic.genetic.Individual;
import ai.metaheuristic.genetic.selector.BiasedWheelOperator;

/**
 * @author Yoann Eichelberger - M1 IWOCS
 */
public class QueenGeneticResolver implements Resolver<int[]> {

    private GeneticAlgorithm<QueenIndividual> genetic;

    /**
     * Construit une instance d'un resolver d'un problème à N-Reines avec
     * l'algorithme génétique.
     * 
     * @param size nombre de reines.
     */
    public QueenGeneticResolver(int size) {
        this.genetic = new GeneticAlgorithm<QueenIndividual>(10000, QueenIndividual[]::new,
                new BiasedWheelOperator<>(fit -> fit + 0.1), 0.15f, 0.05f, 0.01f, 0.00f) {
            @Override
            public QueenIndividual generate() {
                return new QueenIndividual(size);
            }

            @Override
            public QueenIndividual mutate(QueenIndividual individual) {
                int n = individual.queens.length;
                if (n <= 1) {
                    return individual;
                }
                int a = (int) (Math.floor(Math.random() * n));
                int b = (int) (a + 1 + Math.floor(Math.random() * n)) % n;
                int[] copy = Arrays.copyOf(individual.queens, n);
                App.permutate(copy, a, b);
                return new QueenIndividual(copy);
            }

            @Override
            public QueenIndividual crossover(QueenIndividual p1, QueenIndividual p2) {
                int n = size;
                if (p2.lines == null) {
                    p2.lines = new int[n];
                    for (int i = 0; i < n; i++) {
                        p2.lines[p2.queens[i]] = i;
                    }
                }
                int[] child = Arrays.copyOf(p2.queens, n);
                int i1 = (int) Math.floor(Math.random() * n);
                int i2 = p2.lines[p1.queens[i1]];
                App.permutate(child, i1, i2);
                return new QueenIndividual(child);
            }
        };
    }

    /**
     * Recherche une solution.
     * 
     * @return le tableau de reines répondant au problème.
     */
    @Override
    public int[] search() {
        return this.genetic.geneticAlgorithm().queens;
    }

    /**
     * Représente une configuration du placement des reines.
     */
    private static class QueenIndividual extends Individual {

        private int[] queens;
        private int[] lines;
        private double fitness;

        /**
         * Construit un échiquier et place des reines aléatoirement.
         */
        private QueenIndividual(int size) {
            this(QueenIndividual.build(size));
        }

        /**
         * Construit un échiquier à partir d'un autre échiquier.
         */
        private QueenIndividual(int[] queens) {
            this.queens = queens;
            this.lines = null;
            this.fitness = -1;
        }

        /**
         * {@inheritdoc}
         */
        @Override
        public int length() {
            return this.queens.length;
        }

        /**
         * {@inheritdoc}
         */
        @Override
        public double fitness() {
            if (this.fitness == -1) {
                this.fitness = 0;
                for (int i = 1; i < this.queens.length; i++) {
                    if (!App.isTaken(this.queens, i, this.queens[i])) {
                        this.fitness++;
                    }
                }
                this.fitness /= this.queens.length - 1;
            }
            return this.fitness;
        }

        /**
         * Construit un échiquier aléatoirement.
         */
        private static int[] build(int size) {
            int lastIndex = size - 1;
            int[] bag = new int[size];
            for (int i = 0; i < size; i++) {
                bag[i] = i;
            }
            int[] queens = new int[size];
            for (int i = 0; i < size; i++) {
                int j = (int) (Math.random() * lastIndex);
                queens[i] = bag[j];
                bag[j] = bag[lastIndex--];
            }
            return queens;
        }
    }
}
